#include "app-window.h"
#include "rei-json/Object.h"
#include "rei-json/json.h"
#include "selenite/selenite.h"
#include "slint_string.h"
#include "slint_timer.h"
#include "slint.h"
#include <cstdlib>
#include <filesystem>
#include <fstream>
#include <iostream>
#include <ostream>
#include <print>
#include <string>

class CountdownState {
	public:
		CountdownState(slint::ComponentHandle<ui::AppWindow> ui) : ui_(ui) {
			reset();
			ui->on_start_stop([&]{
				if (sessionStep() == ui::SessionStep::Setup || sessionStep() == ui::SessionStep::Finished) {
					reset();
					goNextStep();
					start();
				} else if (sessionStep() == ui::SessionStep::Focus || sessionStep() == ui::SessionStep::Break) {
					toggleCountdown();
				}
			});
		}

		ui::SessionStep sessionStep() const {
			return ui_->global<ui::State>().get_sessions_step();
		}

		int currentSession() const {
			return ui_->global<ui::State>().get_current_session();
		}

		void setCurrentSession(int newCurrentSession) const {
			ui_->global<ui::State>().set_current_session(newCurrentSession);
		}

		int maxSessions() const {
			return ui_->global<ui::State>().get_max_session_count();
		}

		void setSessionStep(ui::SessionStep newSessionStep) {
			ui_->global<ui::State>().set_sessions_step(newSessionStep);
		}

		ui::CountdownStatus countdownStatus() const {
			return ui_->global<ui::State>().get_countdown_status();
		}

		void setCountdownStatus(ui::CountdownStatus newCountdownStatus) {
			ui_->global<ui::State>().set_countdown_status(newCountdownStatus);
		}

		void setCountdown(int newCountdownValue) {
			ui_->global<ui::State>().set_countdown(newCountdownValue);
		}

		int countdown() const {
			return ui_->global<ui::State>().get_countdown();
		}

		void toggleCountdown() {
			if (countdownStatus() == ui::CountdownStatus::Running) {
				pause();
			} else if (countdownStatus() == ui::CountdownStatus::Paused || countdownStatus() == ui::CountdownStatus::NotStarted) {
				start();
			}
		}

		void goNextStep() {
			if (sessionStep() == ui::SessionStep::Setup) {
				setSessionStep(ui::SessionStep::Focus);
				setCountdownStatus(ui::CountdownStatus::NotStarted);
				setCountdown(ui_->global<ui::State>().get_focus_countdown_duration());
				setCurrentSession(currentSession() + 1);
			} else if (sessionStep() == ui::SessionStep::Focus) {
				setSessionStep(ui::SessionStep::Break);
				setCountdownStatus(ui::CountdownStatus::NotStarted);
				setCountdown(ui_->global<ui::State>().get_break_countdown_duration());
			} else if (sessionStep() == ui::SessionStep::Break) {
				setCurrentSession(currentSession() + 1);
				if (currentSession() < maxSessions()) {
					setSessionStep(ui::SessionStep::Focus);
					setCountdownStatus(ui::CountdownStatus::NotStarted);
					setCountdown(ui_->global<ui::State>().get_focus_countdown_duration());
				} else {
					setSessionStep(ui::SessionStep::Finished);
				}
			}
		}

		void start() {
			setCountdownStatus(ui::CountdownStatus::Running);
			timer_.start(slint::TimerMode::Repeated, std::chrono::milliseconds(1000), [&]{
				onTimerTick(&timer_, ui_);
			});
		}

		void pause() {
			setCountdownStatus(ui::CountdownStatus::Paused);
			timer_.stop();
		}

		void reset() {
			timer_.stop();
			setSessionStep(ui::SessionStep::Setup);
			setCountdownStatus(ui::CountdownStatus::NotStarted);
			setCountdown(ui_->global<ui::State>().get_focus_countdown_duration());
			setCurrentSession(0);
		}

		void onTimerTick(slint::Timer* timer, slint::ComponentHandle<ui::AppWindow> ui) {
			int countdown = ui->global<ui::State>().get_countdown();
			countdown--;
			ui->global<ui::State>().set_countdown(countdown);

			

			if (countdown == 0) {
				timer->stop();
				ui->global<ui::State>().set_countdown_status(ui::CountdownStatus::NotStarted); 
				std::string timerEndingScript = std::string(ui->global<ui::State>().get_timer_ending_script());
				if (timerEndingScript.empty()) {
					timerEndingScript = "notify-send -t 0 \"Pomodoro\" \"timer ended!\" && "
										"gst-play-1.0 /usr/share/sounds/freedesktop/stereo/complete.oga &";
				}
				system(timerEndingScript.c_str());
				goNextStep();
			}
		}

	private:

	slint::ComponentHandle<ui::AppWindow> ui_;
	slint::Timer timer_;
};

struct Config {
	std::string timerEndingScript = "";
	int focusDuration = 1500;
	int breakDuration = 300;
	int sessionCount = 4;
};

Config loadJsonConfig(const std::string& configPath) {

	std::ifstream file(configPath);
	Config config{};
	if (!file) {
		std::cerr << "Cant open config file: " << configPath << std::endl;
		return config;
	}
	std::stringstream buffer;
	buffer << file.rdbuf();
	auto jsonStr = buffer.str();
	file.close();

	auto jsonConfig = std::get<rei::json::JsonObject>(rei::json::parse(jsonStr));

	if (jsonConfig["timerEndingScript"].isString()) {
		config.timerEndingScript = jsonConfig["timerEndingScript"].asString();
	}
	if (jsonConfig["focusDuration"].isNumber()) {
		config.focusDuration = jsonConfig["focusDuration"].asNumber();
	}
	if (jsonConfig["breakDuration"].isNumber()) {
		config.breakDuration = jsonConfig["breakDuration"].asNumber();
	}
	if (jsonConfig["sessionCount"].isNumber()) {
		config.sessionCount = jsonConfig["sessionCount"].asNumber();
	}
	return config;
}

void saveJsonConfig(const std::string& configPath, const Config& config) {

	std::ofstream file(configPath);
	if (!file) {
		std::cerr << "Cant open config file: " << configPath << std::endl;
		return;
	}
	rei::json::JsonObject jsonConfig;
	jsonConfig.set("timerEndingScript", config.timerEndingScript);
	jsonConfig.set("focusDuration", config.focusDuration);
	jsonConfig.set("breakDuration", config.breakDuration);
	jsonConfig.set("sessionCount", config.sessionCount);
	file << rei::json::toString(jsonConfig);
	file.close();
}



int main(int argc, char **argv)
{
	std::filesystem::create_directory(std::string(getenv("HOME")) + "/.config/focus");
	const auto configPath = std::string(getenv("HOME")) + "/.config/focus/config.json";

	slint::ComponentHandle<ui::AppWindow> ui = ui::AppWindow::create();

	Config config = loadJsonConfig(configPath);
	ui->global<ui::State>().set_timer_ending_script(slint::SharedString(config.timerEndingScript));
	ui->global<ui::State>().set_focus_countdown_duration(config.focusDuration);
	ui->global<ui::State>().set_break_countdown_duration(config.breakDuration);
	ui->global<ui::State>().set_max_session_count(config.sessionCount);

	CountdownState countdown(ui);

	const auto palettePath = std::string(getenv("HOME")) + "/.config/evalyte/theme.json";
	const auto palette = selenite::parseJson(palettePath);
	if (palette.has_value()) {
		setSelenitePalette<slint::Color, selenite::Color>(ui->global<ui::Palette>(), palette.value());
	}

	ui->global<ui::State>().on_config_changed([&] {
		countdown.reset();
		config.timerEndingScript = ui->global<ui::State>().get_timer_ending_script();
		config.focusDuration = ui->global<ui::State>().get_focus_countdown_duration();
		config.breakDuration = ui->global<ui::State>().get_break_countdown_duration();
		config.sessionCount = ui->global<ui::State>().get_max_session_count();
		saveJsonConfig(configPath, config);
	});

    ui->run();
    return 0;
}
